@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003-2013 The Music Player Daemon Project
+ * Copyright (C) 2003-2014 The Music Player Daemon Project
  * http://www.musicpd.org
  *
  * This program is free software; you can redistribute it and/or modify
@@ -20,19 +20,18 @@
 #include "config.h"
 #include "StickerCommands.hxx"
 #include "SongPrint.hxx"
-#include "DatabaseLock.hxx"
-#include "DatabasePlugin.hxx"
-#include "DatabaseGlue.hxx"
-#include "DatabaseSimple.hxx"
-#include "SongSticker.hxx"
-#include "StickerPrint.hxx"
-#include "StickerDatabase.hxx"
+#include "db/Interface.hxx"
+#include "db/DatabaseGlue.hxx"
+#include "sticker/SongSticker.hxx"
+#include "sticker/StickerPrint.hxx"
+#include "sticker/StickerDatabase.hxx"
 #include "CommandError.hxx"
 #include "protocol/Result.hxx"
+#include "client/Client.hxx"
+#include "Partition.hxx"
+#include "Instance.hxx"
 #include "util/Error.hxx"
 
-#include <glib.h>
-
 #include <string.h>
 
 struct sticker_song_find_data {
@@ -41,7 +40,7 @@ struct sticker_song_find_data {
 };
 
 static void
-sticker_song_find_print_cb(Song &song, const char *value,
+sticker_song_find_print_cb(const LightSong &song, const char *value,
 			   void *user_data)
 {
 	struct sticker_song_find_data *data =
@@ -52,20 +51,20 @@ sticker_song_find_print_cb(Song &song, const char *value,
 }
 
 static CommandResult
-handle_sticker_song(Client &client, int argc, char *argv[])
+handle_sticker_song(Client &client, unsigned argc, char *argv[])
 {
 	Error error;
-	const Database *db = GetDatabase(error);
+	const Database *db = client.GetDatabase(error);
 	if (db == nullptr)
 		return print_error(client, error);
 
 	/* get song song_id key */
 	if (argc == 5 && strcmp(argv[1], "get") == 0) {
-		Song *song = db->GetSong(argv[3], error);
+		const LightSong *song = db->GetSong(argv[3], error);
 		if (song == nullptr)
 			return print_error(client, error);
 
-		const auto value = sticker_song_get_value(song, argv[4]);
+		const auto value = sticker_song_get_value(*song, argv[4]);
 		db->ReturnSong(song);
 		if (value.empty()) {
 			command_error(client, ACK_ERROR_NO_EXIST,
@@ -78,11 +77,11 @@ handle_sticker_song(Client &client, int argc, char *argv[])
 		return CommandResult::OK;
 	/* list song song_id */
 	} else if (argc == 4 && strcmp(argv[1], "list") == 0) {
-		Song *song = db->GetSong(argv[3], error);
+		const LightSong *song = db->GetSong(argv[3], error);
 		if (song == nullptr)
 			return print_error(client, error);
 
-		sticker *sticker = sticker_song_get(song);
+		sticker *sticker = sticker_song_get(*song);
 		db->ReturnSong(song);
 		if (sticker) {
 			sticker_print(client, *sticker);
@@ -92,11 +91,11 @@ handle_sticker_song(Client &client, int argc, char *argv[])
 		return CommandResult::OK;
 	/* set song song_id id key */
 	} else if (argc == 6 && strcmp(argv[1], "set") == 0) {
-		Song *song = db->GetSong(argv[3], error);
+		const LightSong *song = db->GetSong(argv[3], error);
 		if (song == nullptr)
 			return print_error(client, error);
 
-		bool ret = sticker_song_set_value(song, argv[4], argv[5]);
+		bool ret = sticker_song_set_value(*song, argv[4], argv[5]);
 		db->ReturnSong(song);
 		if (!ret) {
 			command_error(client, ACK_ERROR_SYSTEM,
@@ -108,13 +107,13 @@ handle_sticker_song(Client &client, int argc, char *argv[])
 	/* delete song song_id [key] */
 	} else if ((argc == 4 || argc == 5) &&
 		   strcmp(argv[1], "delete") == 0) {
-		Song *song = db->GetSong(argv[3], error);
+		const LightSong *song = db->GetSong(argv[3], error);
 		if (song == nullptr)
 			return print_error(client, error);
 
 		bool ret = argc == 4
-			? sticker_song_delete(song)
-			: sticker_song_delete_value(song, argv[4]);
+			? sticker_song_delete(*song)
+			: sticker_song_delete_value(*song, argv[4]);
 		db->ReturnSong(song);
 		if (!ret) {
 			command_error(client, ACK_ERROR_SYSTEM,
@@ -126,24 +125,17 @@ handle_sticker_song(Client &client, int argc, char *argv[])
 	/* find song dir key */
 	} else if (argc == 5 && strcmp(argv[1], "find") == 0) {
 		/* "sticker find song a/directory name" */
+
+		const char *const base_uri = argv[3];
+
 		bool success;
 		struct sticker_song_find_data data = {
 			client,
 			argv[4],
 		};
 
-		db_lock();
-		Directory *directory = db_get_directory(argv[3]);
-		if (directory == nullptr) {
-			db_unlock();
-			command_error(client, ACK_ERROR_NO_EXIST,
-				      "no such directory");
-			return CommandResult::ERROR;
-		}
-
-		success = sticker_song_find(*directory, data.name,
+		success = sticker_song_find(*db, base_uri, data.name,
 					    sticker_song_find_print_cb, &data);
-		db_unlock();
 		if (!success) {
 			command_error(client, ACK_ERROR_SYSTEM,
 				      "failed to set search sticker database");
@@ -158,7 +150,7 @@ handle_sticker_song(Client &client, int argc, char *argv[])
 }
 
 CommandResult
-handle_sticker(Client &client, int argc, char *argv[])
+handle_sticker(Client &client, unsigned argc, char *argv[])
 {
 	assert(argc >= 4);
 
