@@ -53,10 +53,11 @@ public:
 		children.emplace_back(name, filter);
 	}
 
-	virtual AudioFormat Open(AudioFormat &af, Error &error) override;
-	virtual void Close();
-	virtual ConstBuffer<void> FilterPCM(ConstBuffer<void> src,
-					    Error &error);
+	/* virtual methods from class Filter */
+	AudioFormat Open(AudioFormat &af, Error &error) override;
+	void Close() override;
+	ConstBuffer<void> FilterPCM(ConstBuffer<void> src,
+					    Error &error) override;
 
 private:
 	/**
